/*
*   Title : Mühle game
*   Author: JanR
*   Date  : 29.08.2016
*/


// Game init function, works on a html5 canvas and a message box element
function muehle(canvasElement, messageBox) {

  function lg(msg) {
    if (window.console) {
      window.console.log(msg);
    }
  }

  // Creates and populates the game object and create required shortcuts to
  // common elements as well as design based values for the game
  function initGame(canvasElement, messageBox, game) {

    // The canvas we are working with
    game.canvas = document.getElementById(canvasElement);

    if (game.canvas == null) {
      console.log('Could not find canvas to work on.')
      return;
    }

    // The element which displays information to the player like win or lose
    game.msgBox = document.getElementById(messageBox);

    // The drawing context of the canvas element and setting up a general font
    // size
    game.context = game.canvas.getContext('2d');
    game.context.font = "normal 12px sans-serif";

    // If true, field indexes are rendered
    game.renderFieldLabels = true;

    // If true, field move options are listed
    game.renderMoveLabels = false;

    // If fieldlabels are offf and this is turned on, mills are drawing
    // with special eye-candy
    game.renderSpecialEffect = false;

    // How many times the animation should blink/loop for placement
    // or on removal to visually inform the player
    game.animationBlinks = 2;

    // The increase or decrease each stepping in the animation
    game.animationStep = 0.04;

    // The board background color of the game board
    game.boardColor = 'rgba(255, 220, 120, 1)';

    // The canvas width and height, shortcuts
    game.canvasHeight = game.canvas.height;
    game.canvasWidth = game.canvas.width;

    // State holder of some internally used data for the game to run
    // Mouse cords X and Y
    game.mouseX = 0;
    game.mouseY = 0;

    // How many time did an animation flash
    game.animationFlashes = 0;

    // Is the game animating currently?
    game.isAnimating = false;

    // Buffer to restore a game message for with AI display
    game.previousMessage = "";

    // Where do we move from?
    game.moveSource = null;

    // Which is the last field moved to?
    game.lastMoveField = null;

    // Can any player remove a token?
    game.canRemoveToken = false;

    // Who is the player with the last mill?
    game.lastPlayerMill = 0;

    // Can any player select mills for removal?
    game.highlightMills = false;

    // Is the game over?
    game.finished = false;

    // Game data
    // The fields of the game
    game.fields = [];

    // The current player counted with 0
    game.currentPlayer = 0;

    // The current AI player set to 2 which means black, 1 meaning white
    game.aiPlayer = 2;



    // The spacing of one square to another
    game.rectSpacing = 70;

    // The initial width and height of a square line
    game.initialLineWidth = game.canvasWidth - 80;
    game.initialLineHeight = game.canvasHeight - 80;

    // AI related data
    // Is the AI calculating?
    game.aiCalculating = false;

    // In which stage, for display only, is the AI - just visual so
    game.aiStage = 0;

    // Holder for a intervall where aiStage is displayed in game.msgBox element
    game.AiCalculateInterval = null;

    //--------------------------------------------------------------------------

    // Overview ofg player stones in a 2d array and how many tokens a player
    // currently has in the game
    game.playertokens = {
      'player1': [],
      'player1Count': 0,
      'player2': [],
      'player2Count': 0
    }

    // Populate and init the game fields by quad and index
    for (var quad = 0; quad < 3; ++quad) {
      for (var i = 0; i < 8; ++i) {

        //var random = Math.floor(Math.random() * 3);
        var field = new Object({
          'index': i,                     // Index 0-7
          'quad': quad,                   // Rect number 0-2
          'fieldNumber': i + (8 * quad),  // Real field index
          'state': 0,                     // 0 = Empty, 1 = White, 2 = Black
          'isMill': false,                // Is a mill?
          'highlight': false,             // Is highlighted
          'alpha': 1                      // Alpha value for animation
        });

        game.fields.push(field);
      }
    }

    // Init the game with the "set token phase"
    game.setTokenPhase = true;

    // Set the game set tokens to zero
    game.setTokens = 0;

    // Display a message to the player to place a piece
    displayPlacePiece();

    return true;
  }

  // Resets the game to its initital state so the game routine
  // doenst init every object again, we reset them instead
  this.resetGame = function() {
    // If the game is animating and values will be changed by functions
    // return from reset
    if (game.isAnimating) {
      return;
    }

    // Cleanup the AI calculate message display
    if (game.AiCalculateInterval != null) {
      clearInterval(game.AiCalculateInterval)
      game.AiCalculateInterval = null;
    }

    // Reset the game values
    game.finished = false;

    game.fields = [];
    game.currentPlayer = 0;
    game.aiPlayer = 2;

    game.aiCalculating = false;
    game.aiStage = 0;
    game.AiCalculateInterval = null;

    game.moveSource = null;
    game.lastMoveField = null;

    game.canRemoveToken = false;
    game.lastPlayerMill = 0;
    game.highlightMills = false;

    for (var quad = 0; quad < 3; ++quad) {
      for (var i = 0; i < 8; ++i) {

        var field = new Object({
          'index': i,
          'quad': quad,
          'fieldNumber': i + (8 * quad),
          'state': 0,
          'isMill': false,
          'highlight': false,
          'alpha': 1
        });

        game.fields.push(field);
      }
    }

    game.playertokens = {
      'player1': [],
      'player1Count': 0,
      'player2': [],
      'player2Count': 0
    }

    game.setTokenPhase = true;
    game.setTokens = 0;

    displayPlacePiece();
  }

  //----------------------------------------------------------------------------

  function setMouseGameCoords(event) {
    // Keep track of the mouse cursor position so we can check for mouse overs
    // Please note this is only in Firefox valid and IE
    // There is no browser detection done
    game.mouseX = event.pageX - event.target.offsetLeft;
    game.mouseY = event.pageY - event.target.offsetTop;
  }

  //----------------------------------------------------------------------------

  function drawFields() {
    // Render the game square/quads and game fields

    // Start drawing at 40, 40 and lineWidth and lineHeight at initial length
    var x = 40;
    var y = 40;
    var lineWidth = game.initialLineWidth;
    var lineHeight = game.initialLineHeight;

    for (var i = 0; i < 3; i++) {
      game.context.fillStyle = 'rgba(255,0,0, 1)';
      game.context.strokeStyle = 'rgba(220,50,0, 1)';
      game.context.lineWidth = 6;
      game.context.strokeRect(x, y, lineWidth, lineHeight);

      // Render the field, which is a bit more complexe reusing the current
      // x and y
      renderField(i, 1, x, y, game.highlightMills)
      renderField(i, 2, x + Math.round(lineWidth / 2), y, game.highlightMills)
      renderField(i, 3, x + lineWidth, y, game.highlightMills)

      renderField(i, 8, x, y + Math.round(lineHeight / 2), game.highlightMills)
      renderField(i, 4, x + lineWidth, y + Math.round(lineHeight / 2), game.highlightMills)

      renderField(i, 7, x, y + lineHeight, game.highlightMills)
      renderField(i, 5, x + lineWidth, y + lineHeight, game.highlightMills)
      renderField(i, 6, x + Math.round(lineWidth / 2), y + lineHeight, game.highlightMills)

      // After drawing the quads, increase x and y to grow fruther, while
      // reducing the width and height of the drawing boundaries
      x += game.rectSpacing;
      y += game.rectSpacing;
      lineWidth -= game.rectSpacing * 2;
      lineHeight -= game.rectSpacing * 2;
    }
  }

  //----------------------------------------------------------------------------

  function drawGroundBoard() {
    // Draws the ground board inner going connections using basic canvas methods
    var x = 40;
    var y = 40;
    var lineWidth = game.initialLineWidth;
    var lineHeight = game.initialLineHeight;

    game.context.fillStyle = 'rgba(220, 50,0, 1)';
    game.context.fillRect(x + (lineWidth / 2)- 3, y, 6, (game.rectSpacing * 2));
    game.context.fillRect(x + (lineWidth / 2)- 3, lineHeight - (game.rectSpacing * 2) + y, 6, game.rectSpacing * 2);
    game.context.fillRect(x, y + (lineHeight / 2) - 3, (game.rectSpacing * 2), 6);
    game.context.fillRect(x + lineWidth - (game.rectSpacing * 2), y + (lineHeight / 2) - 3, (game.rectSpacing * 2), 6);
  }

  //----------------------------------------------------------------------------

  function renderField(quadNumber, fieldNumber, x, y, highlightMills) {

    // Renders the fields, mills, magic effects when mills are closed


    var fieldIndex = (fieldNumber + (8 * quadNumber))-1;

    // Draw the base outline
    if (game.fields[fieldIndex].state != 0 && game.lastPlayerMill != 0 && game.fields[fieldIndex].state != game.lastPlayerMill) {

      // Defaults to white
      var fillColor = 'rgba(255, 255,255, 1)';

      // If it is a mill token...
      if (game.fields[fieldIndex].isMill) {
        // ... and we are allowed to highlight mills (for removal)
        if (game.highlightMills) {
          fillColor = 'rgba(255, 0, 0, 1)';
        }
      } else {
        // Is not a token in a mill.
        fillColor = 'rgba(255, 0, 0, 1)';
      }

      game.context.fillStyle = fillColor;
    } else if (game.fields[fieldIndex].highlight) {
      // Is the field set hovered
      game.context.fillStyle = 'rgba(255, 255, 0, 1)';
    } else {
      // Defaults to white
      game.context.fillStyle = 'rgba(255, 255,255, 1)';
    }

    game.context.strokeStyle = 'rgba(128,128,128, 1)';
    game.context.lineWidth = 3;

    game.context.beginPath();
    game.context.moveTo(x, y);
    game.context.arc(x, y, 12, 0, (Math.PI / 180) * 360, true);

    //--------------------------------------------------------------------------

    // Draw any highlighting for mouse over
    if (!game.aiCalculating && !game.isAnimating && !game.finished && game.context.isPointInPath(game.mouseX, game.mouseY)) {
      // Set the active mousefield to the hovered field which is used later on
      game.mouseField = game.fields[fieldIndex];

      if (game.mouseField.state == (game.currentPlayer + 1)) {
        game.context.fillStyle = 'rgba(255, 180, 0, 1)';
        if (game.mouseField.highlight) {
          game.context.fillStyle = 'rgba(255, 255, 0, 1)';
        }
      }
    }

    game.context.stroke();
    game.context.fill();
    game.context.closePath();

    //--------------------------------------------------------------------------


    // Draw the game tokens by player color if not empty
    if (game.fields[fieldIndex].state != 0) {
      if (game.fields[fieldIndex].state == 1) {
        game.context.fillStyle = 'rgba(200, 200, 200, '+ game.fields[fieldIndex].alpha.toString() + ')';
      } else {
        game.context.fillStyle = 'rgba(80, 80, 80, ' + game.fields[fieldIndex].alpha.toString() + ')';
      }
      game.context.strokeStyle = 'rgba(0, 0, 0, ' + game.fields[fieldIndex].alpha.toString() + ')';

      game.context.beginPath();
      game.context.lineWidth = 4;
      game.context.arc(x, y, 8, 0, (Math.PI / 180) * 360, true);
      game.context.stroke();
      game.context.fill();
      game.context.closePath();
    }

    if (game.fields[fieldIndex].isMill) {
      game.context.fillStyle = 'rgba(255, 255, 255, ' + game.fields[fieldIndex].alpha.toString() + ')';
      game.context.lineWidth = 3;
      game.context.beginPath();
      game.context.arc(x, y, 4, 0, (Math.PI / 180) * 360, true);
      game.context.stroke();
      game.context.fill();
      game.context.closePath();

      if (game.renderSpecialEffect) {
        game.context.fillStyle = 'rgba(0, 0, 100, ' + game.fields[fieldIndex].alpha.toString() + ')';
        game.context.fillText("5", x-4, y+5);
      }
    }

    // Draw optional field labels
    if (!game.renderSpecialEffect && game.renderFieldLabels) {
        game.context.fillStyle = 'rgba(0, 0, 100, 1)';
        game.context.fillText(fieldIndex.toString(), x-(game.context.measureText(fieldIndex.toString()).width / 2), y+5);
    }
  }

  //----------------------------------------------------------------------------

  function findMills() {
    // Function to detect if there are any mills on the board and flag them as
    // mills.
    // Also set the "game.lastPlayerMill" variable based on "game.lastMoveField"
    // to determine if the last move created a mill

    for (var quadNumber = 0; quadNumber < 3; ++quadNumber) {
      var currentPlayer = 3;
      var playerStones = 0;
      var hasWrapped = false;

      for (var i = 0; i < 9; ++i) {
        if (i == 8) {
          i = 0;
          hasWrapped = true;
        } else {
          game.fields[i + (8 * quadNumber)].isMill = false;
        }

        var fieldIndex = i + (8 * quadNumber)
        var field = game.fields[fieldIndex];


        if (field.state == 0) {
          playerStones = 0;
          currentPlayer = 3;
        } else {
          // continue her to check for player stones
          if (currentPlayer == field.state) {
            ++playerStones;

            if (playerStones >= 3) {
              if (field.index % 2 == 0) {
                playerStones = 1;
              }

              var mill1Index = null;
              var mill2Index = null;
              var mill3Index = null;

              if (field.index != 0 && field.index % 2 == 0) {
                mill1Index = fieldIndex-2;
                mill2Index = fieldIndex-1;
                mill3Index = fieldIndex;
              } else if (field.index == 0) {
                mill1Index = fieldIndex+7;
                mill2Index = fieldIndex+6;
                mill3Index = fieldIndex;
              } else {
                continue;
              }

              game.fields[mill1Index].isMill = true;
              game.fields[mill2Index].isMill = true;
              game.fields[mill3Index].isMill = true;

              if (game.lastMoveField != null && (game.lastMoveField == game.fields[mill1Index] || game.lastMoveField == game.fields[mill2Index] || game.lastMoveField == game.fields[mill3Index])) {
                game.lastPlayerMill = game.lastMoveField.state;
              }
            }

          } else {
            currentPlayer = field.state;
            playerStones = 1;
          }
        }

        if (hasWrapped) {
          break;
        }
      }
    }

    for (i = 1; i < 8; i += 2) {
      var fieldQuad1 = i;
      var fieldQuad2 = i + 8;
      var fieldQuad3 = i + 16;

      if (game.fields[fieldQuad1].state != 0) {
        var currentPlayer = game.fields[fieldQuad1].state;
        if (game.fields[fieldQuad2].state == currentPlayer && game.fields[fieldQuad3].state == currentPlayer) {
          game.fields[fieldQuad1].isMill = true;
          game.fields[fieldQuad2].isMill = true;
          game.fields[fieldQuad3].isMill = true;

          if (game.lastMoveField != null && (game.lastMoveField == game.fields[fieldQuad1] || game.lastMoveField == game.fields[fieldQuad2] || game.lastMoveField == game.fields[fieldQuad3])) {
            game.lastPlayerMill = game.lastMoveField.state;
          }
        }
      }
    }

    //--------------------------------------------------------------------------

    if (game.lastMoveField != null && game.lastPlayerMill != 0) {
      // Recreate the player token set
      setPlayerTokens();

      if (!game.setTokenPhase) {
        // Check for losing/winning conditions
        checkGameState();
      }

      if (!game.finished) {
        // If the game is ongoing, show a message to remove a stone
        game.msgBox.innerHTML = "Bitte entfernen Sie einen "+ (game.lastMoveField.state == 2 ? "weißen" : "schwarzen") + " Spielstein!"

        // Set the game.highlightMills variable
        setHighlightMills(game.lastPlayerMill);

        // Set the game to canRemoveToken
        game.canRemoveToken = true;
      }

    }
  }

  //----------------------------------------------------------------------------

  function setHighlightMills(currentPlayer) {
    // Determines if any player token is not inside a mill
    // and sets "game.highlightMills" to true or false

    var playerTokens = [];
    if (currentPlayer == 1) {
      playerTokens = game.playertokens.player2;
    } else {
      playerTokens = game.playertokens.player1;
    }

    for (var i = 0; i < playerTokens.length; ++i) {
      if (playerTokens[i] != 0) {
        if (!game.fields[i].isMill) {
          game.highlightMills = false;
          return;
        }
      }
    }
    game.highlightMills = true;
  }

  function checkClick(event) {
    // Main mouse click handler

    if (game.isAnimating || game.aiCalculating) {
      return;
    }

    if (game.mouseField == null) {
      return;
    }

    if (game.setTokenPhase && !game.canRemoveToken) {
      // If we are in the setTokenPhase and we cant remove a token
      // place a token
      placePiece(false)
    } else if (!game.canRemoveToken) {
      // If we cannot remove any token, we are moving
      movePiece(false);
    } else if (game.lastPlayerMill != 0) {
      // If there was a mill, we can remove a piece on click
      removePiece(false);
    }
  }

  //----------------------------------------------------------------------------
  function displayPlacePiece() {
    // Displays whos turn it is to place a token on the game board
    // using "game.msgBox"

    if (game.setTokens < 18) { // TODO: set 18 here!
      game.msgBox.innerHTML = (game.setTokens % 2 == 0 ? 'WEIß' : 'SCHWARZ') + ' IST AM ZUG EINEN STEIN ZU PLAZIEREN.<br/>Stein ' + (Math.ceil((game.setTokens+1)/2)).toString() + ' von 9';
      game.previousMessage = game.msgBox.innerHTML;
      return;
    }

    game.currentPlayer = 1;
    game.setTokenPhase = false;
    switchTurn();
  }

  function switchTurn() {
    // Switches the control from one player to another

    game.currentPlayer = !game.currentPlayer;

    if (!game.setTokenPhase) {
      if (game.currentPlayer == 0) {
        game.msgBox.innerHTML = "WEIß IST AM ZUG";
      } else {
        game.msgBox.innerHTML = "SCHWARZ IST AM ZUG";
      }

      checkGameState();
      game.previousMessage = game.msgBox.innerHTML;
    } else {
      displayPlacePiece();
    }

    game.mouseField = null;
    game.moveSource = null;

    // Give control to the AI player
    if ((game.currentPlayer + 1) == game.aiPlayer && !game.finished) {
      game.aiCalculating = true;
      game.aiStage = 0;
      game.AiCalculateInterval = setInterval(displayAiCalculateMessage, 500);
      setTimeout(calculateAIturn, 800);
    }
  }

  //----------------------------------------------------------------------------

  function placePiece(isAi) {

    // Place a piece on the board
    // "isAi" tells if this is the computer or human player

    // Search all available fields and get the information if the player
    // can set a stone there

    var possibleMoves = searchPossibleMoves(0, true);
    for (var i = 0; i < possibleMoves.length; ++i) {
      if (possibleMoves[i].index == game.mouseField.index && possibleMoves[i].quad == game.mouseField.quad) {
        ++game.setTokens;

        game.mouseField.state = ((game.setTokens+1) % 2) + 1;
        animateField(game.mouseField, 1);
        game.lastMoveField = game.mouseField;

        setPlayerTokens();
        findMills();

        if (!game.canRemoveToken) {
          if (!game.finished && !isAi) {
            switchTurn();
          }
        }

        break;
      }
    }
  }

  //----------------------------------------------------------------------------

  function removePiece(isAi) {
    // Remove a piece from the board
    // "isAi" tells if this is the computer or human player

    if (game.mouseField.state != 0 && game.mouseField.state != game.lastPlayerMill) {

      // Finalize the removal after the animation has been done to also switch
      // control to another player and check win/lose conditions
      function doProcessRemove() {
        if (game.isAnimating) {
          // Loop the request to the next animation frame
          window.requestAnimationFrame(doProcessRemove);
        } else {
          game.lastMoveField = null;
          game.lastPlayerMill = 0;
          game.canRemoveToken = false;

          setPlayerTokens();
          findMills();

          if (!game.setTokenPhase) {
            checkGameState();
          }

          if (!game.finished && !isAi) {
            switchTurn();
          }
        }
      }

      if (game.mouseField.isMill && !game.highlightMills) {
        return;
      }

      // Call the function to animate the game.mouseField to 0 alpha
      animateField(game.mouseField, 0);
      // Initial call to doProcessRemove, could also be used as a callback so...
      doProcessRemove();
    }
  }

  //----------------------------------------------------------------------------

  function movePiece(isAi) {
    // Moves a pieces from one point to another
    // +  figures out the selection, unselect, select change of a players token
    //    and shows possible moves for the selected field.
    // "isAi" tells if this is the computer or human player

    if (game.moveSource == null && game.mouseField.state == 0) {
      return;
    }

    if (game.moveSource == null) {
      if (game.mouseField.state != (game.currentPlayer + 1)) {
        return;
      }
      game.moveSource = game.mouseField;
      game.moveSource.highlight = true;
      return;
    } else if (game.moveSource == game.mouseField) {
      game.moveSource.highlight = false;
      game.moveSource = null;
      return;
    } else if (game.moveSource.state == game.mouseField.state) {
      game.moveSource.highlight = false;
      game.moveSource = null;
      game.moveSource = game.mouseField;
      game.moveSource.highlight = true;
      return;
    }

    var canJump = false;
    if (game.moveSource.state == 1 && game.playertokens.player1Count == 3) {
      canJump = true;
    } else if (game.moveSource.state == 2 && game.playertokens.player2Count == 3) {
      canJump = true;
    }

    var possibleMoves = searchPossibleMoves(game.moveSource.index + (8 * game.moveSource.quad), canJump);

    // Function to end the move animation and switch the control to another player
    // here used as a call back to "animateMove(...)"

    function finishProcessMove(sourceField, targetField) {
      game.moveSource = sourceField;
      game.mouseField = targetField;

      game.mouseField.alpha = 1;
      game.mouseField.state = game.moveSource.state;
      game.lastMoveField = game.mouseField;

      game.moveSource.state = 0;
      game.moveSource.highlight = false;
      game.moveSource = null;

      setPlayerTokens();

      findMills();
      checkGameState();

      if (!game.finished && !game.canRemoveToken && !isAi) {
        switchTurn();
      }
    }

    for (var i = 0; i < possibleMoves.length; ++i) {
      if (possibleMoves[i].index == game.mouseField.index && possibleMoves[i].quad == game.mouseField.quad) {
        // Animate moves animates the movement and call finisheProcessMove
        animateMove(game.moveSource, game.mouseField, finishProcessMove);
        break;
      }
    }
  }

  function animateMove(sourceField, targetField, callback) {

    // Function to draw the animation of a token from one field to
    // another at a constant rate

    // sourceField is the field we are starting at

    // targetField is the field we want to animate to

    // callback is the function which becomes called after the animation
    // finished

    //--------------------------------------------------------------------------

    // Set active that we are animating
    game.isAnimating = true;

    // Begin the calculation for start field to target field coordinates

    var x = 40;
    var y = 40;

    var startX = sourceField.quad * game.rectSpacing;
    var startY = sourceField.quad * game.rectSpacing;

    var targetX = targetField.quad * game.rectSpacing;
    var targetY = targetField.quad * game.rectSpacing;

    var startLineWidth = game.initialLineWidth - (startX * 2);
    var startLineHeight = game.initialLineHeight - (startY * 2);

    var targetLineWidth = game.initialLineWidth - (targetX * 2);
    var targetLineHeight = game.initialLineHeight - (targetY * 2);

    if (sourceField.index != 0) {
      if (sourceField.index == 1 || sourceField.index == 5) {
        startX += Math.round(startLineWidth / 2);
      } else if (sourceField.index > 1 && sourceField.index < 5) {
        startX += startLineWidth;
      }

      if (sourceField.index == 3 || sourceField.index == 7) {
        startY += Math.round(startLineHeight / 2);
      } else if (sourceField.index > 3 && sourceField.index < 7) {
        startY += startLineHeight;
      }
    }

    if (targetField.index != 0) {
      if (targetField.index == 1 || targetField.index == 5) {
        targetX += Math.round(targetLineWidth / 2);
      } else if (targetField.index > 1 && targetField.index < 5) {
        targetX += targetLineWidth;
      }

      if (targetField.index == 3 || targetField.index == 7) {
        targetY += Math.round(targetLineHeight / 2);
      } else if (targetField.index > 3 && targetField.index < 7) {
        targetY += targetLineHeight;
      }
    }


    startX += x;
    startY += y;
    targetX += x;
    targetY += y;

    var distanceX = startX - targetX;
    var distanceY = startY - targetY;

    var stepX = 0;
    var stepY = 0;
    if (distanceX != 0) {
      stepX = distanceX * 0.04
    }

    if (distanceY != 0) {
      stepY = distanceY * 0.04;
    }

    distanceX = Math.abs(startX - targetX);
    distanceY = Math.abs(startY - targetY);

    var state = sourceField.state;

    //--------------------------------------------------------------------------

    // Function to do the actual movement

    function doAnimation() {
      if ((distanceX + distanceY) > 0) {

        if (distanceX != 0) {
          distanceX -= Math.abs(stepX);
        }

        if (distanceY != 0) {
          distanceY -= Math.abs(stepY);
        }

        if ((distanceX + distanceY) <= 0) {
          // The animation is finished, call callback function
          game.isAnimating = false;
          callback(sourceField, targetField);
          return;
        } else {
          startX -= stepX;
          startY -= stepY;

          game.context.fillStyle = state == 1 ? 'rgba(200, 200, 200, 1)' : 'rgba(80, 80, 80, 1)';
          game.context.strokeStyle = 'rgba(0, 0, 0, 1)';
          game.context.lineWidth = 4;

          game.context.moveTo(startX, startY);
          game.context.beginPath();
          game.context.arc(startX, startY, 8, 0, (Math.PI / 180) * 360, true);
          game.context.stroke();
          game.context.fill();
          game.context.closePath();
          window.requestAnimationFrame(doAnimation);
          return;
        }
      }
    }

    doAnimation();
  }

  //----------------------------------------------------------------------------

  function animateField(targetField, targetAlpha) {

    // Animates a target field to a target alpha
    // uses animationFlashes internally so multiple fields could be animated
    // simultaniously

    game.isAnimating = true;

    if (targetAlpha != 0) {
      targetField.alpha = 0;
    } else {
      targetField.alpha = 1;
    }

    var animationFlashes = 0;

    function doAnimation() {
      if (targetAlpha != 0) {
        targetField.alpha += game.animationStep;

        if (targetField.alpha >= 1) {
          ++animationFlashes;

          if (animationFlashes < game.animationBlinks) {
            targetField.alpha = 0;
            window.requestAnimationFrame(doAnimation);
            return;
          }

          game.isAnimating = false;
          targetField.alpha = 1;
          return;
        }
      } else {
        targetField.alpha -= game.animationStep;

        if (targetField.alpha <= 0) {
          ++animationFlashes;

          if (animationFlashes < game.animationBlinks) {
            targetField.alpha = 1;
            window.requestAnimationFrame(doAnimation);
            return;
          }

          game.isAnimating = false;
          targetField.isMill = false;
          targetField.state = 0;
          targetField.alpha = 1;

          setPlayerTokens();
          if (!game.setTokenPhase) {
            checkGameState();
          }

          return;
        }
      }

      window.requestAnimationFrame(doAnimation);
    }

    doAnimation();
  }

  //----------------------------------------------------------------------------

  function searchPossibleMoves(index, canJump) {

    // Searches all possible moves for a
    // given field by game.fields direct index: index
    // and has a switch if the player can jump with his tokens

    if (canJump) {
      // If we can jump, all fields are neighbours
      var neighbours = game.fields;
    } else {
      // Determine the direct neighbours
      var neighbours = [];
      var field = game.fields[index];

      if (field.index != 0) {
        if ((field.index + 1) % 4 == 0) {
          if (field.quad == 0) {
            neighbours = [game.fields[index + 8], game.fields[index - 1], field.index != 7 ? game.fields[index + 1] : game.fields[index - 7]];
          } else if (field.quad == 1) {
            neighbours = [game.fields[index - 8], game.fields[index - 1], game.fields[index + 8], field.index != 7 ? game.fields[index + 1] : game.fields[index - 7]];
          } else {
            neighbours = [game.fields[index - 8], game.fields[index - 1], field.index != 7 ? game.fields[index + 1] : game.fields[index - 7]];
          }
        } else if ((field.index + 1) % 2 == 0) {
          if (field.quad == 0) {
            neighbours = [game.fields[index + 8], game.fields[index - 1], game.fields[index + 1]];
          } else if (field.quad == 1) {
            neighbours = [game.fields[index - 8], game.fields[index + 8], game.fields[index - 1], game.fields[index + 1]];
          } else {
            neighbours = [game.fields[index - 8], game.fields[index - 1], game.fields[index + 1]];
          }
        } else {
          neighbours = [game.fields[index - 1], game.fields[index + 1]];
        }
      } else {
        neighbours = [game.fields[7 + (8 * field.quad)], game.fields[field.index + 1 + (8 * field.quad)]];
      }
    }

    // Collect all possible moves based on empty neighbours
    var possibleMoves = [];
    for (var i = 0; i < neighbours.length; ++i) {
      if (neighbours[i].state == 0) {
        possibleMoves.push(neighbours[i]);
      }
    }

    return possibleMoves;
  }


  function renderPossibleMoves() {
    // Renders all possible moves (in orange) where the player can acutally move

    // Return if a player can remove a token or we are in setTokenPhase
    if (game.canRemoveToken || game.setTokenPhase) {
      return;
    }

    var canJump = false;
    var srcField = {};

    if (game.moveSource == null) {
      srcField = game.mouseField;
      // If we dont have any active selected token, we check the mouse field as
      // source
    } else {
      // If we have a field selected, use this as source field for option display
      srcField = game.moveSource;
    }

    if (srcField.state == 1 && game.playertokens.player1Count == 3) {
      canJump = true;
    } else if (srcField.state == 2 && game.playertokens.player2Count == 3) {
      canJump = true;
    }

    possibleMoves = searchPossibleMoves(srcField.index + (8 * (srcField.quad)), canJump);

    // Check how many move options the player has, for display
    var possibleMoveCount = 0;

    for (var i = 0; i < possibleMoves.length; ++i) {
      var neighbour = possibleMoves[i];

      // Figure out the coords for the current possible move
      ++possibleMoveCount;

      var x = 40;
      var y = 40;
      var lineWidth = game.initialLineWidth - (neighbour.quad * (game.rectSpacing * 2));
      var lineHeight = game.initialLineHeight - (neighbour.quad * (game.rectSpacing * 2));
      if (!canJump) {
        x += (neighbour.quad * game.rectSpacing);

        if (neighbour.index != 0) {
          if ((srcField.index + 1) % 4 == 0) {
            if (srcField.index == 3) {
              x += lineWidth;
            }
          } else if ((srcField.index + 1) % 2 == 0) {
            if (neighbour.index < 3) {
              x += lineWidth / (3 - neighbour.index);
            } else if (neighbour.index < 5) {
              x += lineWidth;
            } else if (neighbour.index == 5) {
              x += lineWidth / (7 - neighbour.index);
            }
          } else if (neighbour.index != 7) {
            if (neighbour.index < 3 || neighbour.index > 4) {
              x += lineWidth / 2;
            } else if (neighbour.index > 1) {
              x += lineWidth;
            }
          }
        }

        y += (neighbour.quad * game.rectSpacing);

        if (neighbour.index != 0) {
          if ((srcField.index + 1) % 4 == 0) {
            if (neighbour.index == 3 || neighbour.index == 7) {
              y += lineHeight / 2;
            } else if (neighbour.index > 3) {
              y += lineHeight;
            }
          } else if (srcField.index == 5) {
            y += lineHeight;
          } else if (neighbour.index == 3  || neighbour.index == 7) {
            y += lineHeight / 2;
          } else if (neighbour.index > 3) {
            y += lineHeight;
          }
        }
      } else {
        x += (neighbour.quad * game.rectSpacing);
        if (neighbour.index != 0 && neighbour.index < 6) {
          if (neighbour.index == 1 || neighbour.index == 5) {
            x += lineWidth / 2;
          } else {
            x += lineWidth;
          }
        }

        y += (neighbour.quad * game.rectSpacing);
        if (neighbour.index != 0 && neighbour.index > 2) {
          if (neighbour.index == 3 || neighbour.index == 7) {
            y += lineHeight / 2;
          } else {
            y += lineHeight;
          }
        }
      }

      // Draw the possible move option
      game.context.beginPath();
      game.context.fillStyle = 'rgba(255, 180, 0, 1)';
      game.context.moveTo(x, y);
      game.context.arc(x, y, 12, 0, (Math.PI / 180) * 360, true);
      game.context.fill();
      game.context.closePath();

      if (game.renderMoveLabels) {
        game.context.fillStyle = 'rgba(0, 0, 0, 1)';
        game.context.fillText(possibleMoveCount.toString(), x - (game.context.measureText(possibleMoveCount.toString()).width/2), y+5);
      }
    }
  }

  //----------------------------------------------------------------------------

  function setPlayerTokens() {
    // Keep a internal state of all player tokens for each game field
    // and set the playertokens count accordingly

    // This is used after removal for example or before checking the game state
    game.playertokens.player1 = [];
    game.playertokens.player1Count = 0;

    game.playertokens.player2 = [];
    game.playertokens.player2Count = 0;

    for (var i = 0; i < game.fields.length; i++) {
      if (game.fields[i].state == 0) {
        game.playertokens.player1.push(0);
        game.playertokens.player2.push(0);
      } else if (game.fields[i].state == 1) {
        ++game.playertokens.player1Count;
        game.playertokens.player1.push(1);
        game.playertokens.player2.push(0);
      } else {
        ++game.playertokens.player2Count;
        game.playertokens.player1.push(0);
        game.playertokens.player2.push(1);
      }
    }
  }

  //----------------------------------------------------------------------------

  function checkGameState() {
    // Checks the game state for winning/losing conditions and displays
    // a message about the game state to the player(s)

    // Check for less then 2 stones
    if (game.playertokens.player1Count <= 2) {
      game.msgBox.innerHTML = "SCHWARZ GEWINNT DIE PARTIE<br/>"+game.playertokens.player1Count.toString()+" weiße Steine übrig.";
      game.finished = true;
      return;
    } else if (game.playertokens.player2Count <= 2) {
      game.msgBox.innerHTML = "WEIß GEWINNT DIE PARTIE!<br/>"+game.playertokens.player2Count.toString()+" schwarze Steine übrig.";
      game.finished = true;
      return;
    }

    // Check for out of moves
    for (var i = 0; i < 2; ++i) {
      var playerTokens = [];

      if (i == 0) {
        if (game.playertokens.player1Count == 3) {
          // Player can jump
          continue;
        }

        playerTokens = game.playertokens.player1;
      } else {
        if (game.playertokens.player2Count == 3) {
          // Player can jump
          continue;
        }

        playerTokens = game.playertokens.player2;
      }

      var hasPossibleMove = false;
      for (var j = 0; j < playerTokens.length; ++j) {
        if (playerTokens[j] != 0) {
          var possibleMoves = searchPossibleMoves(j, false);

          if (possibleMoves.length != 0) {
            hasPossibleMove = true;
            break;
          }
        }
      }

      if (!hasPossibleMove) {
        if (i == 0) {
          game.msgBox.innerHTML = "SCHWARZ GEWINNT DIE PARTIE!<br/>Weiß kann keinen Zug mehr machen.";
        } else {
          game.msgBox.innerHTML = "WEIß GEWINNT DIE PARTIE!<br/>Schwarz kann keinen Zug mehr machen.";
        }
        game.finished = true;
        return;
      }
    }

  }

  //----------------------------------------------------------------------------

  function displayAiCalculateMessage() {
    // Simply display "the AI is calculating message" to let the player know
    // that the AI is currently working

    game.msgBox.innerHTML = game.previousMessage + '<br/>' + ('...').substr(0, game.aiStage) + (game.currentPlayer == 0 ? "Weiß" : "Schwarz") + ' berechnet den Zug' + ('...').substr(0, game.aiStage);
    if (game.aiStage == 3) {
      game.aiStage = 0;
    } else {
      ++game.aiStage;
    }
  }

  //----------------------------------------------------------------------------

  function gameloop() {
    // The main game Loop

    game.context.fillStyle = game.boardColor;
    game.context.fillRect(0, 0, game.canvasWidth, game.canvasHeight);
    game.mouseField = null;

    // Draw the ground board
    drawGroundBoard();

    // Draw the fields
    drawFields();

    //console.log(game.mouseField);

    if (!game.isAnimating) {
      if (game.moveSource != null || (game.mouseField != null && game.mouseField.state == (game.currentPlayer + 1))) {
        // Render possible moves if a token is selected
        renderPossibleMoves();
      }
    }

    window.requestAnimationFrame(gameloop);
  }

  //----------------------------------------------------------------------------

  function calculateAIturn() {
    // Calculate the AI turn

    // Helper function if a field index
    function fieldInList(fieldList, fieldNumber) {

      for (var i = 0; i < fieldList.length; ++i) {
        if (fieldNumber == fieldList[i]) {
          return true;
        }
      }

      return false;
    }

    // Helper function to sort mill, opponenten, own, available fields
    // by score
    function sortByScore(a ,b) {
      return a[0] < b[0];
    }

    // Calculate the fields score for a given field
    function getFieldScore(fieldNumber, mode) {
      var field = game.fields[fieldNumber];

      if (mode != 0) {
        var opponentPlayer = game.currentPlayer + 1;
        var currentPlayer = !game.currentPlayer + 1;
      } else {
        var opponentPlayer = !game.currentPlayer + 1;
        var currentPlayer = game.currentPlayer + 1;
      }

      var neighbours = [];
      var indirectNeighbours = [];
      var score = 0;

      if (field.index == 0) {
        score = -5;
        neighbours = [fieldNumber + 7, fieldNumber + 6, fieldNumber + 1, fieldNumber + 2];
        if (field.quad == 0) {
          indirectNeighbours = [fieldNumber + 15, fieldNumber + 9];
        } else if (field.quad == 1) {
          indirectNeighbours = [fieldNumber - 7, fieldNumber + 1, fieldNumber + 9, fieldNumber + 15];
        } else {
          indirectNeighbours = [fieldNumber - 1, fieldNumber - 8, fieldNumber + 3, fieldNumber + 5];
        }
      } else if (field.index % 2 == 0) {
        if (field.index != 6) {
          neighbours = [fieldNumber - 2, fieldNumber - 1, fieldNumber + 1, fieldNumber + 2];
        } else {
          neighbours = [fieldNumber - 2, fieldNumber - 1, fieldNumber + 1, fieldNumber - 6];
        }

        if (field.quad == 0) {
          score = -5;
          indirectNeighbours = [fieldNumber + 7, fieldNumber + 9, fieldNumber + 17];
        } else if (field.quad == 1) {
          score = 0;
          indirectNeighbours = [fieldNumber - 7, fieldNumber - 9, fieldNumber + 7, fieldNumber + 9];
        } else {
          score = -3;
          indirectNeighbours = [fieldNumber - 15, fieldNumber - 7, fieldNumber - 9, fieldNumber - 3, fieldNumber - 5];
        }

      } else if ((field.index + 1) % 8 == 0) {
        if (field.quad == 0) {
          score = -5
          neighbours = [fieldNumber - 7, fieldNumber - 1, fieldNumber + 8, fieldNumber + 16];
          indirectNeighbours = [fieldNumber + 1, fieldNumber + 7];
        } else if (field.quad == 1) {
          score = 7
          neighbours = [fieldNumber - 7, fieldNumber - 1, fieldNumber - 8, fieldNumber + 8];
          indirectNeighbours = [fieldNumber - 15, fieldNumber - 9, fieldNumber - 6, fieldNumber - 2, fieldNumber + 1, fieldNumber + 7];
        } else {
          score = -2
          neighbours = [fieldNumber - 7, fieldNumber - 1, fieldNumber - 8, fieldNumber - 16];
          indirectNeighbours = [fieldNumber - 15, fieldNumber - 9, fieldNumber - 6, fieldNumber - 2, fieldNumber - 14, fieldNumber - 10];
        }
      } else if ((field.index + 1) % 4 == 0) {
        if (field.quad == 0) {
          score = -5
          neighbours = [fieldNumber - 1, fieldNumber + 1, fieldNumber + 8, fieldNumber + 16];
          indirectNeighbours = [fieldNumber + 7, fieldNumber + 9];
        } else if (field.quad == 1) {
          score = 7
          neighbours = [fieldNumber - 1, fieldNumber + 1, fieldNumber - 8, fieldNumber + 8];
          indirectNeighbours = [fieldNumber - 9, fieldNumber - 7, fieldNumber - 2, fieldNumber + 2, fieldNumber + 7, fieldNumber + 9];
        } else {
          score = -2
          neighbours = [fieldNumber - 1, fieldNumber + 1, fieldNumber - 8, fieldNumber - 16];
          indirectNeighbours = [fieldNumber - 9, fieldNumber - 7, fieldNumber - 2, fieldNumber + 2, fieldNumber - 10, fieldNumber - 6];
        }
      } else if ((field.index + 1) % 2 == 0) {
        if (field.quad == 0) {
          score = -5
          neighbours = [fieldNumber - 1, fieldNumber + 1, fieldNumber + 8, fieldNumber + 16];
          indirectNeighbours = [fieldNumber + 7, fieldNumber + 9]
        } else if (field.quad == 1) {
          score = 7
          neighbours = [fieldNumber - 1, fieldNumber + 1, fieldNumber - 8, fieldNumber + 8];
          indirectNeighbours = [fieldNumber - 9, fieldNumber - 7, fieldNumber + 6, fieldNumber + 2, fieldNumber + 7, fieldNumber + 9];
        } else {
          score = -2
          neighbours = [fieldNumber - 1, fieldNumber + 1, fieldNumber - 16, fieldNumber - 8];
          indirectNeighbours = [fieldNumber - 7, fieldNumber - 9, fieldNumber -2, fieldNumber + 2];
        }
      }

      if (mode != 0) {
        //score = 0;
      }

      // TODO: Add a function to check the direct neighbours of a field and deceide if the player blocks itself from moving
      var ownCount = 0;
      var opponentCount = 0;
      var emptyFields = 0;

      for (var i = 0; i < neighbours.length; ++i) {
        if (mode != 0) {
          // mode 1 is used for removal

          if (game.fields[neighbours[i]].state == currentPlayer) {
            score += game.fields[neighbours[i]].isMill ? 35 : 16;
            ++ownCount;
          } else if (game.fields[neighbours[i]].state == 0) {
            score += 8;
            ++emptyFields;
          } else if (game.fields[neighbours[i]].state == opponentPlayer) {
            score += game.fields[neighbours[i]].isMill ? 11 : 44;
            ++opponentCount;
          }
        } else {
          if (game.fields[neighbours[i]].state == currentPlayer) {
            score += game.fields[neighbours[i]].isMill ? 15 : 42;
            ++ownCount;
          } else if (game.fields[neighbours[i]].state == 0) {
            score += 14;
            ++emptyFields;
          } else if (game.fields[neighbours[i]].state == opponentPlayer) {
            score += game.fields[neighbours[i]].isMill ? 46 : 18;
            ++opponentCount;
          }
        }
      }

      if (mode == 0 && opponentCount >= 2 && ownCount != 1 && emptyFields >= 2) {
        score += 90;
      } else if (opponentCount >= 2 && emptyFields >= 2) {
        score += 90;
      }

      for (var i = 0; i < indirectNeighbours.length; ++i) {
        if (mode != 0) {
          if (game.fields[indirectNeighbours[i]].state == currentPlayer) {
            score += game.fields[indirectNeighbours[i]].isMill ? 28 : 12;
          } else if (game.fields[indirectNeighbours[i]].state == 0) {
            score += 1;
          } else if (game.fields[indirectNeighbours[i]].state == opponentPlayer) {
            score += game.fields[indirectNeighbours[i]].isMill ? 16 : 2;
          }
        } else {
          if (game.fields[indirectNeighbours[i]].state == currentPlayer) {
            score += game.fields[indirectNeighbours[i]].isMill ? 30 : 24;
          } else if (game.fields[indirectNeighbours[i]].state == 0) {
            score += 1;
          } else if (game.fields[indirectNeighbours[i]].state == opponentPlayer) {
            score += game.fields[indirectNeighbours[i]].isMill ? 38 : 26;
          }
        }
      }

      return score;
    }

    // Provided a 2d array with two indexes, return the one with the best
    // mill by score

    function getStrongestMillField(millPositions) {
      var targetMill = [];
      var highestScore = 0;

      if (millPositions[1].length != 0) {
        targetMill = millPositions[1][Math.floor(Math.random() * millPositions[1].length)];
        highestScore = targetMill[0];
        millField = targetMill[1];
        millPosition = targetMill[2];

        for (var i = 0; i < millPositions[1].length; ++i) {
          if (millPositions[1][i][0] > highestScore) {
            targetMill = millPositions[1][i];
            highestScore = targetMill[0];
            millField = targetMill[1];
            millPosition = targetMill[2];
          }
        }
      }

      if (millPositions[0].length != 0) {
        if (highestScore < millPositions[0][0]) {
          targetMill = millPositions[0][0];
          highestScore = targetMill[0];
          millField = targetMill[1];
          millPosition = targetMill[2];
        }

        for (var i = 0; i < millPositions[0].length; ++i) {
          if (millPositions[0][i][0] > highestScore) {
            targetMill = millPositions[0][i];
            highestScore = targetMill[0];
            millField = targetMill[1];
            millPosition = targetMill[2];
          }
        }
      }

      return targetMill;
    }

    //--------------------------------------------------------------------------

    function findMillPositions(availableFields, ownFields, opponentFields) {

      // Find possible mills for a player, providing the available, own and
      // opponentfields

      // From all available fields
      // The opponenent fields are counted negative and are excluded from choice
      // Own fields are counted positive and are excluded from choice

      // They are then sorted and return as 2d Array with the field:
      // [Score, game.fields index, [positions]]


      var millPositions = [[],[]];
      var mill1 = [0, 0];
      var mill2 = [];

      for (var i = 0; i < availableFields.length; ++i) {
        var fieldNumber = availableFields[i];
        var score = 0;

        if (fieldNumber == 0 || fieldNumber % 8 == 0) {
          mill1 = [fieldNumber, fieldNumber + 6, fieldNumber + 7];
          mill2 = [fieldNumber, fieldNumber + 1, fieldNumber + 2];
        } else if (fieldNumber % 2 == 0) {
          mill1 = [fieldNumber, fieldNumber - 2, fieldNumber - 1];
          mill2 = [fieldNumber, fieldNumber + 1, fieldNumber + 2];
        } else if ((fieldNumber + 1) % 8 == 0) {
          mill1 = [fieldNumber, fieldNumber - 7, fieldNumber - 1];
          mill2 = [fieldNumber, fieldNumber - 16, fieldNumber - 8, fieldNumber + 8, fieldNumber + 16];
        } else if ((fieldNumber + 1) % 2 == 0) {
          mill1 = [fieldNumber, fieldNumber - 1, fieldNumber + 1];
          mill2 = [fieldNumber, fieldNumber - 16, fieldNumber - 8, fieldNumber + 8, fieldNumber + 16];
        }

        var isOccupied = false;
        var mill1Filtered = [];
        score = 0;

        for (var j = 0; j < mill1.length; ++j) {
          if (fieldInList(opponentFields, mill1[j])) {
            --score;
          } else {
            if (game.fields[mill1[j]].isMill) {
              ++score;
            }

            if (fieldInList(ownFields, mill1[j])) {
              ++score;
            } else {
              mill1Filtered.push(mill1[j]);
            }
          }
        }

        millPositions[0].push([score, fieldNumber, mill1Filtered]);

        var mill2Filtered = [];
        isOccupied = false;
        score = 0;

        for (var j = 0; j < mill2.length; ++j) {
          if (mill2[j] >= 0 && mill2[j] < game.fields.length) {
            if (fieldInList(opponentFields, mill2[j])) {
              --score;
            } else {
              if (game.fields[mill2[j]].isMill) {
                ++score;
              }

              if (fieldInList(ownFields, mill2[j])) {
               ++score;
             } else {
               mill2Filtered.push(mill2[j]);
             }
            }
          }
        }

        millPositions[1].push([score, fieldNumber, mill2Filtered]);
      }

      millPositions[0].sort(sortByScore);
      millPositions[1].sort(sortByScore);

      return millPositions;
    }

    //--------------------------------------------------------------------------

    // With all functions declared the AI starts its turn

    // aiLevel = 0 (Random based)
    // aiLevel = 1 (Some tactics)

    var aiLevel = 1;

    // Data holders
    var ownFields = [];
    var opponentFields = [];
    var availableFields = [];

    // Filter the game fields
    for (var i = 0; i < game.fields.length; ++i) {
      if (game.fields[i].state == 0) {
        // Push to empty fields
        availableFields.push(i);
      } else if (game.fields[i].state == game.aiPlayer) {
        // Push to own fields
        ownFields.push(i);
      } else {
        // Push to opponent fields
        opponentFields.push(i);
      }
    }


    // If we are in the set token phase
    if (game.setTokenPhase) {
      var targetFieldIndex = -1;

      if (aiLevel == 0) {
        targetFieldIndex = availableFields[Math.floor(Math.random() * availableFields.length)];
      } else if (aiLevel == 1) {
        var millPositions = findMillPositions(availableFields, ownFields, opponentFields); // [ MillScore, [Field1, Field2...] ]
        var targetMill = getStrongestMillField(millPositions);

        var availableFieldScores = [];
        var aiFieldScores = [];
        var opponentFieldScores = [];

        for (var i = 0; i < availableFields.length; ++i) {
          availableFieldScores.push( [getFieldScore(availableFields[i], 0), availableFields[i]] );
        }

        for (var i = 0; i < ownFields.length; ++i) {
          aiFieldScores.push( [getFieldScore(ownFields[i], 0), ownFields[i]] );
        }

        for (var i = 0; i < opponentFields.length; ++i) {
          opponentFieldScores.push( [getFieldScore(opponentFields[i], 1), opponentFields[i]] );
        }

        availableFieldScores.sort(sortByScore);
        aiFieldScores.sort(sortByScore);
        opponentFieldScores.sort(sortByScore);


        lg("# START -------------------------------#");

        for (var i = 0; i < availableFieldScores.length; ++i) {
          lg("#"+availableFieldScores[i][1].toString()+" : " + availableFieldScores[i][0].toString());
        }

        lg("=======================================");
        /*
        for (var i = 0; i < aiFieldScores.length; ++i) {
          lg("#"+aiFieldScores[i][1].toString()+" : " + aiFieldScores[i][0].toString());
        }

        lg("=======================================");
        */

        for (var i = 0; i < opponentFieldScores.length; ++i) {
          lg("#"+opponentFieldScores[i][1].toString()+" : " + opponentFieldScores[i][0].toString());
        }

        lg("=======================================");
        /*
        lg("\nMILL 1");

        for (var i = 0; i < millPositions[0].length; ++i) {
          lg("#"+millPositions[0][i][1].toString()+" : " + millPositions[0][i][0].toString());
        }

        lg("\nMILL 2");

        for (var i = 0; i < millPositions[1].length; ++i) {
          lg("#"+millPositions[1][i][1].toString()+" : " + millPositions[1][i][0].toString());
        }

        lg("=======================================");
        */

        lg("# END ---------------------------------#");

        var randomChance = 0;

        if (targetMill.length != 0) {
          if (targetMill[0] > 1) {
            randomChance = 0.8;
          } else if (targetMill[1] == 1) {
            randomChance = 0.7;
          }

          if (availableFieldScores[0][0] > 70 && availableFieldScores[0][0] < 200) {
            randomChance -= 0.25;
          } else if (availableFieldScores[0][0] >= 200) {
            randomChance = 0;
          }
        }

        if (game.setTokens < game.aiPlayer) {
          targetFieldIndex = availableFieldScores[Math.floor(Math.random() * Math.floor(availableFieldScores.length/2.5))][1];
          //targetFieldIndex = availableFields[Math.floor(Math.random() * availableFields.length)];
        } else {
          if (randomChance != 0 && Math.random() < randomChance) {
            targetFieldIndex = targetMill[2][Math.floor(Math.random() * targetMill[2].length)];
          } else {
            targetFieldIndex = availableFieldScores[0][1];
          }
        }
      }

      game.mouseField = game.fields[targetFieldIndex];
      placePiece(true);
    }

    //--------------------------------------------------------------------------

    if (!game.setTokenPhase) {
      var availableMoveOptions = [];
      var targetPlayer = game.currentPlayer + 1;
      var targetFieldIndex = -1;
      var sourceFieldIndex = 0;

      //------------------------------------------------------------------------

      var possibleMoves = [];
      var canJump = targetPlayer == 1 ? (game.playertokens.player1Count <= 3 ? true : false) : (game.playertokens.player2Count <= 3 ? true : false);

      for (var i = 0; i < game.fields.length; ++i) {
        if (game.fields[i].state == targetPlayer) {
          if (canJump) {
            possibleMoves = searchPossibleMoves(0, canJump);
          } else {
            possibleMoves = searchPossibleMoves(game.fields[i].index + (8 * game.fields[i].quad), false);
          }

          if (possibleMoves.length != 0) {
            availableMoveOptions.push([i, possibleMoves]);
          }
        }
      }

      //------------------------------------------------------------------------
      if (aiLevel == 0) {
        sourceFieldIndex = Math.floor(Math.random() * availableMoveOptions.length);
        targetFieldIndex = availableMoveOptions[sourceFieldIndex][0];
        targetField = availableMoveOptions[sourceFieldIndex][1][Math.floor(Math.random() * availableMoveOptions[sourceFieldIndex][1].length)];
      } else {
        sourceFieldIndex = Math.floor(Math.random() * availableMoveOptions.length);
        targetFieldIndex = availableMoveOptions[sourceFieldIndex][0];
        targetField = availableMoveOptions[sourceFieldIndex][1][Math.floor(Math.random() * availableMoveOptions[sourceFieldIndex][1].length)];
      }

      //------------------------------------------------------------------------

      game.moveSource = game.fields[targetFieldIndex];
      game.mouseField = targetField;
      movePiece(true);
    }



    function finishMove() {
      // Switch control to the other player after all animations are finished
      // moving/placing animationStep

      // and call a piece for removal in case the option occured

      if (game.isAnimating) {
        // While the game is animating recall the function
        window.requestAnimationFrame(finishMove);
        return;
      }

      if (game.canRemoveToken) {
        var targetFields = [];
        var targetPlayer = !game.currentPlayer + 1;
        var targetFieldIndex = 0;

        for (var i = 0; i < game.fields.length; ++i) {
          if (game.fields[i].state == targetPlayer && (!game.fields[i].isMill || game.highlightMills)) {
            targetFields.push(i);
          }
        }

        if (aiLevel == 0) {
          targetFieldIndex = targetFields[Math.floor(Math.random() * targetFields.length)];
        } else {
          var fieldScores = [];
          for (var i = 0; i < targetFields.length; ++i) {
            fieldScores.push( [getFieldScore(targetFields[i], 1), targetFields[i]] );
          }

          fieldScores.sort(sortByScore);

          lg("# START REMOVAL -----------------------#");
          for (var i = 0; i < fieldScores.length; ++i) {
            lg("#"+fieldScores[i][1].toString()+" : " + fieldScores[i][0].toString());
          }

          lg("# END REMOVAL -------------------------#");

          if (!game.highlightMills) {
            for (var i = 0; i < fieldScores.length; ++i) {
              if (!game.fields[fieldScores[i][1]].isMill) {
                targetFieldIndex = fieldScores[i][1];
                break;
              }
            }
          } else {
            targetFieldIndex = fieldScores[0][1];
          }
        }

        game.mouseField = game.fields[targetFieldIndex];
        removePiece(true);
      }

      clearInterval(game.AiCalculateInterval);
      game.aiCalculating = false;
      game.AiCalculateInterval = null;
      switchTurn();
    }


    finishMove();
  }

  //----------------------------------------------------------------------------


  // Init start ----------------------------------------------------------------
  var game = {};
  if (!initGame(canvasElement, messageBox, game)) {
    console.log("Error initializing the game.");
    return;
  }

  window.requestAnimationFrame = window.requestAnimationFrame || window.mozRequestAnimationFrame || window.webkitRequestAnimationFrame || window.oRequestAnimationFrame;
  game.canvas.addEventListener("mousemove", setMouseGameCoords);
  game.canvas.addEventListener("mouseover", setMouseGameCoords);
  game.canvas.addEventListener("click", checkClick);
  gameloop();
}

// Final call to init the game intially
var game = new muehle("playboard", "message");
